from .download import download_zip
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "libs.zip",down_path)


bl_info = {
    "name": "Section Box",
    "description": "Visualization, Sections, Elevations, and more vfxMed",
    "blender": (3, 4, 0),
    "version": (2, 1, 2),
    "category": "3D View",
    "author": "Ruben Messerschmidt",
    "location": "View3d > Sidebar > Section Box",
    "doc_url": "https://section-box.readthedocs.io/en/latest/",
    "version_url": "https://raw.githubusercontent.com/rubenmesserschmidt/Section-Box_docs/main/version",
}

addon_id = __package__


import bpy
import sys
import importlib
import pkgutil

from pathlib import Path
from .translation import translate


##### Load Modules #####


def get_all_submodules(directory):
    return list(iter_submodules(directory, directory.name))


def iter_submodules(path, package_name):
    for name in sorted(iter_submodule_names(path)):
        mod = importlib.import_module("." + name, __package__)
        yield mod


def iter_submodule_names(path, root=""):
    for _, module_name, is_package in pkgutil.iter_modules([str(path)]):
        if is_package:
            if module_name == "standard_presets":
                continue
            sub_path = path / module_name
            sub_root = root + module_name + "."
            yield from iter_submodule_names(sub_path, sub_root)
        else:
            yield root + module_name


def reload_modules():
    for name in reload_list:
        importlib.reload(sys.modules[name])


def get_loaded_modules():
    prefix = __name__ + "."
    return [name for name in sys.modules if name.startswith(prefix)]


if "reload_list" in locals():
    reload_modules()
else:
    modules = get_all_submodules(Path(__file__).parent)
    reload_list = get_loaded_modules()


##### Registration #####


from .support import Support
from .update import UpdateCheck
from .preferences import AddonPreferences
from .dependencies import InstallModulesPopup
from .restart import Restart
from .props.sectionplane_settings import SECTIONBOX_PG_SectionPlaneSettings
from .props.sectionbox_settings import SECTIONBOX_PG_SectionBoxSettings
from .handlers import (
    init,
    store_props,
    update_props,
    convert,
)
from .update import invoke_popup
from .keymaps import register_keymaps, unregister_keymaps
from .icons import register_icons, unregister_icons
from .ops import register_ops, unregister_ops
from .props import register_props, unregister_props
from .ui import register_ui, unregister_ui
from .ui.menus import object_context_menu, prop_context_menu
from .utils.convert_old import convert_keymaps, convert_prefs, convert_info
from .utils.addon import is_addon_enabled, is_addon_installed


# right click menu
class UI_MT_button_context_menu(bpy.types.Menu):
    bl_label = ""

    def draw(self, context):
        pass


class SB_PG_View(bpy.types.PropertyGroup):
    name: bpy.props.StringProperty()
    plane_transforms: bpy.props.FloatVectorProperty(size=6)

    edge_transforms: bpy.props.FloatVectorProperty(size=24)

    invert: bpy.props.BoolProperty()


classes = [
    SECTIONBOX_PG_SectionPlaneSettings,
    SECTIONBOX_PG_SectionBoxSettings,
    SB_PG_View,
    UI_MT_button_context_menu,
    Support,
    UpdateCheck,
    AddonPreferences,
    InstallModulesPopup,
    Restart,
]


def register():
    # check Blender version
    if bpy.app.version < (3, 4, 0):
        print(f"{bl_info['name']} is not compatible with your current Blender version!")
        print(
            f'Upgrade to Blender {str((3, 4, 0)).replace("(", "").replace(")", "")} or later.'
        )
        return

    # classes
    for cls in classes:
        try:
            bpy.utils.register_class(cls)
        except Exception as e:
            print(f" {cls} register failed: {e}")

    # icons
    register_icons()

    # props
    register_props()

    # ops
    register_ops()

    # ui
    register_ui()

    # keymaps
    register_keymaps()

    # menus
    bpy.types.VIEW3D_MT_edit_mesh_context_menu.append(object_context_menu)
    bpy.types.VIEW3D_MT_object_context_menu.append(object_context_menu)
    bpy.types.OUTLINER_MT_collection.append(object_context_menu)
    bpy.types.OUTLINER_MT_object.append(object_context_menu)
    bpy.types.UI_MT_button_context_menu.append(prop_context_menu)

    # handlers
    bpy.app.handlers.load_post.append(init)
    bpy.app.handlers.undo_post.append(init)
    bpy.app.handlers.frame_change_pre.append(store_props)
    bpy.app.handlers.frame_change_post.append(update_props)

    if is_addon_enabled("Section Box"):
        # convert old
        bpy.app.timers.register(lambda: convert_keymaps(), first_interval=0.1)
        bpy.app.timers.register(lambda: convert_prefs(), first_interval=0.1)

    if is_addon_installed("Section Box"):
        bpy.app.timers.register(lambda: convert(None), first_interval=0.2)
        bpy.app.timers.register(lambda: convert_info(), first_interval=0.2)

    bpy.app.handlers.load_post.append(convert)

    # init
    bpy.app.timers.register(lambda: init(None), first_interval=0.1)

    # invoke dependency popup
    bpy.app.timers.register(
        lambda: InstallModulesPopup.invoke_popup("PREFERENCES"),
        first_interval=0.1,
    )

    # update check
    # bpy.app.timers.register(
    #     update.invoke_popup,
    #     first_interval=0.1
    # )
    translate.register()

def unregister():
    # remove draw handler
    dh = bpy.app.driver_namespace.get("sectionbox.draw_handles")
    if dh:
        dh.remove_handler()
        del bpy.app.driver_namespace["sectionbox.draw_handles"]

    # end modals
    bpy.context.window_manager.sectionbox.running_modals = False

    # handlers
    bpy.app.handlers.load_post.remove(init)
    bpy.app.handlers.undo_post.remove(init)
    bpy.app.handlers.frame_change_post.remove(update_props)
    bpy.app.handlers.frame_change_pre.remove(store_props)
    bpy.app.handlers.load_post.remove(convert)

    # menus
    bpy.types.VIEW3D_MT_edit_mesh_context_menu.remove(object_context_menu)
    bpy.types.VIEW3D_MT_object_context_menu.remove(object_context_menu)
    bpy.types.OUTLINER_MT_collection.remove(object_context_menu)
    bpy.types.OUTLINER_MT_object.remove(object_context_menu)
    bpy.types.UI_MT_button_context_menu.remove(prop_context_menu)

    # ui
    unregister_ui()

    # keymaps
    unregister_keymaps()

    # ops
    unregister_ops()

    # props
    unregister_props()

    # icons
    unregister_icons()

    # classes
    for cls in reversed(classes):
        try:
            bpy.utils.unregister_class(cls)
        except Exception as e:
            print(f" {cls} unregister failed: {e}")
            
    translate.unregister()

if __name__ == "__main__":
    register()
