from pathlib import Path
from threading import Thread
from itertools import chain
from contextlib import suppress

import bpy
import pkg_resources

from .ops.popup_base import BLENDAI_OT_PopupBase
from .utils.ui import draw_paragraph, close_dialog
from .utils.system import (
    internet_connection,
    install_module,
    get_missing_modules,
    get_os,
)
from . import addon_id


addon_path = Path(__file__).parent

installing = False


class InstallModulesPopup(BLENDAI_OT_PopupBase, bpy.types.Operator):
    bl_idname = "sectionbox.install_modules_popup"
    bl_description = "Install the additional modules required by this addon"

    modules: bpy.props.StringProperty()
    feature: bpy.props.StringProperty()

    icon = "LINKED"

    _th = None
    _timer = None
    _success = False

    @staticmethod
    def invoke_popup(area_type):
        context = bpy.context
        window = None
        area = None
        for win in context.window_manager.windows:
            for a in win.screen.areas:
                if a.type == area_type:
                    area = a
                    window = win
                    break

            if window:
                break

        if not (window or area):
            return

        with context.temp_override(window=window, area=area):
            operator_id = InstallModulesPopup.bl_idname.split(".")[-1]
            getattr(getattr(bpy.ops, "sectionbox"), operator_id)("INVOKE_DEFAULT")

    def execute(self, context):
        wm = context.window_manager

        self._th = Thread(target=self.install)
        self._th.start()
        self._timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)
        return {"RUNNING_MODAL"}

    def invoke(self, context, event):
        if self.action == "install":
            if not self.check_connection():
                return {"CANCELLED"}

            global installing
            installing = True
            return self.execute(context)

        missing_modules = self.get_missing_modules()
        if not missing_modules:
            return {"CANCELLED"}

        self.missing_modules = missing_modules
        return super().invoke(context, event)

    def get_modules(self):
        requirements = addon_path.joinpath("requirements.txt")
        if requirements.exists():
            return list(pkg_resources.parse_requirements(requirements.open()))

        requirements_common = addon_path.joinpath("requirements_common.txt")
        requirements_os = addon_path.joinpath(f"requirements_{get_os().lower()}.txt")
        if not requirements_common.exists() or not requirements_os.exists():
            return []

        return list(
            chain(
                pkg_resources.parse_requirements(requirements_common.open()),
                pkg_resources.parse_requirements(requirements_os.open()),
            )
        )

    def get_missing_modules(self):
        if self.modules:
            return get_missing_modules(self.modules.split(","))

        modules = self.get_modules()
        return get_missing_modules(modules)

    def modal(self, context, event):
        if event.type == "TIMER":
            if self._th.is_alive():
                self.report({"WARNING"}, message="Installing modules...")
            else:
                self.finish(context)
                return {"FINISHED"}
        return {"RUNNING_MODAL"}

    def finish(self, context):
        if not self._success:
            self.report(
                {"ERROR"},
                "Could not install modules. Please check your internet connection!",
            )
            self.clean(context)
            self.cancel(context)

        self.report({"INFO"}, "Installed modules successfully!")
        self.clean(context)
        getattr(bpy.ops, "sectionbox").restart("INVOKE_DEFAULT")

    def clean(self, context):
        with suppress(Exception):
            context.window_manager.event_timer_remove(self._timer)

        self._timer = None
        self._th = None
        self._success = False

    def cancel(self, context):
        global installing
        if installing:
            return

        if self.get_missing_modules() and not self.feature:
            bpy.ops.preferences.addon_disable(module=addon_id)

    def draw(self, context):
        super().draw(context)
        layout = self.layout
        text = f"{self.feature} requires the following additional modules:"
        # draw_paragraph(
        #     layout,
        #     text=text,
        #     alignment="CENTER",
        # )
        row = layout.row()
        row.alignment = "CENTER"
        box = row.box()
        missing_module_names = [m.project_name for m in self.missing_modules]
        for module_name in missing_module_names:
            col = box.column()
            col.scale_y = 0.8
            row = col.row(align=True)
            row.label(icon="DOT")
            row.label(text=module_name)

        layout.separator(factor=1)

        text = (
            "This is a one time process and can take a moment."
            "\n\nDO NOT leave the preferences window until the installation is complete!"
        )
        # draw_paragraph(
        #     layout,
        #     text=text,
        #     alignment="CENTER",
        # )
        row = layout.row()
        row.active_default = True
        op = row.operator(self.bl_idname, text="Install Modules")
        op.action = "install"
        op.modules = ",".join(missing_module_names)

    def check_connection(self):
        self.report({"INFO"}, "Checking internet connection...")
        if not internet_connection():
            self.report(
                {"ERROR"},
                "Could not connect to server, please check your internet connection!",
            )
            return False

        return True

    def install(self):
        modules = self.modules.split(",")
        for module in modules:
            if not install_module(module):
                self._success = False
                return

        self._success = True
