import bpy
from bpy.types import Operator

import bmesh

from ..utils.ui import expand_layout, split_layout
from ..utils.sectionbox import get_sectionbox


class SECTIONBOX_OT_Export(Operator):
    bl_idname = "sectionbox.export"
    bl_label = "Export"
    bl_options = {"INTERNAL"}

    @classmethod
    def description(cls, context, properties):
        if properties.drawing == "cross_section":
            return "Export cross section"
        elif properties.drawing == "elevation":
            return "Export elevation"

    drawing: bpy.props.StringProperty(options={"HIDDEN"})  # cross_section, elevation
    format: bpy.props.StringProperty(options={"HIDDEN"})  # dxf, mesh
    reset_position: bpy.props.BoolProperty(options={"HIDDEN"})
    clean_up: bpy.props.BoolProperty(
        name="Clean up",
        description="Remove doubles and dissolve superfluous vertices",
        options={"HIDDEN"},
    )
    clean_up_dst: bpy.props.FloatProperty(
        name="Max distance",
        description="Distance threshold for removing doubles",
        subtype="DISTANCE",
        default=0.0001,
        min=0.0,
        max=1.0,
        step=0.001,
        options={"HIDDEN"},
    )
    clean_up_angle: bpy.props.FloatProperty(
        name="Max angle",
        description="Angle threshold for dissolving vertices",
        default=0.0872665,
        subtype="ANGLE",
        min=0.0,
        max=3.14159,
        step=0.0872665,
        options={"HIDDEN"},
    )

    filepath: bpy.props.StringProperty(
        subtype="FILE_PATH",
    )
    filename: bpy.props.StringProperty(
        subtype="FILE_NAME",
    )

    def invoke(self, context, event):
        if self.format in {"dxf"}:
            context.window_manager.fileselect_add(self)
            return {"RUNNING_MODAL"}

        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        split = layout.split()
        lcol = split.column()
        rcol = split.column()
        expand_layout(
            lcol, "Clean Up", 0, context.scene.sectionbox.ui, "expand_clean_up"
        )
        rcol.prop(self, "clean_up", toggle=True, icon="SHADERFX", text="")
        if context.scene.sectionbox.ui.expand_clean_up:
            sub = layout.column(align=True)
            sub.enabled = self.clean_up
            lcol, rcol = split_layout(sub, spacing=2.5)
            lcol.label(text="Doubles")
            rcol.prop(self, "clean_up_dst", text="")

            lcol, rcol = split_layout(sub, spacing=2.5)
            lcol.label(text="Dissolve")
            rcol.prop(self, "clean_up_angle", text="")

    def execute(self, context):
        sb = get_sectionbox(context)
        plane = sb.planes[sb.active_plane]
        name = "Cross Section" if self.drawing == "cross_section" else "Elevation"
        if plane.cross_section:
            cs_obj = plane.cross_section.copy()
            cs_obj.data = plane.cross_section.data.copy()
            cs_bm = bmesh.new()
            cs_bm.from_mesh(cs_obj.data)
            if self.clean_up:
                self._clean_up(context, [cs_bm])

        if self.drawing == "elevation":
            ele_obj = plane.elevation.copy()
            ele_obj.data = plane.elevation.data.copy()
            ele_bm = bmesh.new()
            ele_bm.from_mesh(ele_obj.data)
            if self.clean_up:
                self._clean_up(context, [ele_bm])

        if self.format == "dxf":
            import ezdxf

            if not self.filepath.endswith(".dxf"):
                self.filepath = f"{self.filepath}.dxf"

            doc = ezdxf.new()
            msp = doc.modelspace()
            doc.layers.add(name="Cross Section")
            if plane.cross_section:
                for edge in cs_bm.edges:
                    msp.add_line(
                        edge.verts[0].co.xy,
                        edge.verts[1].co.xy,
                        dxfattribs={"layer": "Cross Section"},
                    )
                cs_bm.free()
                bpy.data.meshes.remove(cs_obj.data)

            if self.drawing == "elevation":
                doc.layers.add(name="Elevation")
                for edge in ele_bm.edges:
                    msp.add_line(
                        edge.verts[0].co.xy,
                        edge.verts[1].co.xy,
                        dxfattribs={"layer": "Elevation"},
                    )
                ele_bm.free()
                bpy.data.meshes.remove(ele_obj.data)

            doc.saveas(self.filepath)
            self.report(
                {"INFO"}, f"{name} exported as {self.format} under {self.filepath}!"
            )

        elif self.format == "mesh":
            obj_name = f"{sb.entry_name} - {plane.object.name[-1]}"
            [o.select_set(False) for o in context.selected_objects]
            if plane.cross_section:
                cs_bm.to_mesh(cs_obj.data)
                context.scene.collection.objects.link(cs_obj)
                cs_obj.name = f"Cross Section - {obj_name}"
                cs_obj.hide_viewport = False
                cs_obj.hide_render = False
                cs_obj.hide_select = False
                cs_obj.select_set(True)
                context.view_layer.objects.active = cs_obj

            if self.drawing == "elevation":
                ele_bm.to_mesh(ele_obj.data)
                context.scene.collection.objects.link(ele_obj)
                ele_obj.name = f"Elevation - {obj_name}"
                ele_obj.hide_viewport = False
                ele_obj.hide_render = False
                ele_obj.hide_select = False
                ele_obj.select_set(True)
                context.view_layer.objects.active = ele_obj

            self.report({"INFO"}, f"{name} exported as {self.format}!")

        return {"FINISHED"}

    def _clean_up(self, context, bms):
        for bm in bms:
            bmesh.ops.remove_doubles(
                bm,
                verts=bm.verts,
                dist=self.clean_up_dst,
            )
            bmesh.ops.dissolve_verts(
                bm,
                verts=[
                    v
                    for v in bm.verts
                    if len(v.link_edges) == 2
                    and v.calc_edge_angle(0) < self.clean_up_angle
                ],
                use_boundary_tear=False,
                use_face_split=False,
            )
