import bpy
from bpy.types import Operator

from bpy_extras import view3d_utils
from mathutils import Vector

from ..utils.preferences import get_prefs
from ..utils.sectionbox import get_sectionbox


class SECTIONBOX_OT_Select(Operator):
    bl_idname = "sectionbox.select"
    bl_label = "Select"
    bl_description = "Select"
    bl_options = {"REGISTER", "UNDO", "INTERNAL"}

    def invoke(self, context, event):
        self.prefs = get_prefs()
        self.key_select = (
            context.window_manager.keyconfigs.user.keymaps["Object Mode"]
            .keymap_items["sectionbox.select"]
            .type
        )
        self.sb = get_sectionbox(context, check_active=True)
        if not self.sb:
            return {"PASS_THROUGH"}

        self.mouse_start = Vector((event.mouse_region_x, event.mouse_region_y))
        if not context.window_manager.sectionbox.edge_select:
            self.plane = self.sb.hit_plane(context, self.mouse_start)
            if self.plane is None:
                return {"PASS_THROUGH"}

            self.opposite_plane = self.sb.opposite_plane(self.plane)
            p_loc = self.plane.location_2d(context)
            op_loc = self.opposite_plane.location_2d(context)
            if p_loc is None or op_loc is None:
                return {"PASS_THROUGH"}

            self.direction_2d = (
                self.opposite_plane.location_2d(context)
                - self.plane.location_2d(context)
            ).normalized()
            self.distance_2d = (
                self.opposite_plane.location_2d(context)
                - self.plane.location_2d(context)
            ).length
            self.start_transform = self.plane.transform
            self.distance = (
                self.opposite_plane.location() - self.plane.location()
            ).length
        else:
            self.plane = self.sb.planes[self.sb.active_plane]
            self.edge_id = self.plane.select_edge(context, self.mouse_start)
            if self.edge_id is None:
                return {"PASS_THROUGH"}
            self.edge = [
                self.sb.object.data.edges[self.sb.object.data.edge_keys.index(ek)]
                for ek in self.sb.object.data.polygons[self.plane.id].edge_keys
            ][self.edge_id]
            self.opposite_edge = self.sb.opposite_edge(self.plane, self.edge)
            edge_location = (
                self.sb.object.data.vertices[self.edge.vertices[0]].co
                + self.sb.object.data.vertices[self.edge.vertices[1]].co
            ) / 2
            opposite_edge_location = (
                self.sb.object.data.vertices[self.opposite_edge.vertices[0]].co
                + self.sb.object.data.vertices[self.opposite_edge.vertices[1]].co
            ) / 2

            edge_location_2d = view3d_utils.location_3d_to_region_2d(
                context.region,
                context.space_data.region_3d,
                self.sb.object.matrix_world @ edge_location,
            )

            opposite_edge_location_2d = view3d_utils.location_3d_to_region_2d(
                context.region,
                context.space_data.region_3d,
                self.sb.object.matrix_world @ opposite_edge_location,
            )
            dir = opposite_edge_location_2d - edge_location_2d
            self.direction_2d = dir.normalized()
            self.distance_2d = dir.length
            self.start_transform = self.plane.edge_transform[self.edge_id]
            self.distance = (opposite_edge_location - edge_location).length

        if self.distance_2d < 0.0001:
            return {"PASS_THROUGH"}

        context.window_manager.modal_handler_add(self)
        if self.prefs.lazy_load:
            self.sb.set_update(False)
        return {"RUNNING_MODAL"}

    def modal(self, context, event):
        # on release
        if event.type == self.key_select and event.value == "RELEASE":
            self.sb.update_cutter(context)
            self.sb.set_update(True)
            return {"FINISHED"}

        # on cancel
        if event.type in {
            "RIGHTMOUSE" if self.key_select == "LEFTMOUSE" else "LEFTMOUSE",
            "ESC",
        }:
            if not context.window_manager.sectionbox.edge_select:
                self.plane.transform = self.start_transform
            else:
                self.plane.edge_transform[self.edge_id] = self.start_transform

            self.sb.update_transform(context)
            self.sb.set_update(True)
            return {"CANCELLED"}

        # on drag
        if event.type == "MOUSEMOVE":
            mouse_current = Vector((event.mouse_region_x, event.mouse_region_y))
            mouse_delta = self.direction_2d.dot(mouse_current - self.mouse_start)
            factor = mouse_delta / self.distance_2d
            if not context.window_manager.sectionbox.edge_select:
                self.plane.transform = (
                    self.start_transform + min(factor, 0.999) * self.distance
                )
            else:
                self.plane.edge_transform[self.edge_id] = (
                    self.start_transform + min(factor, 0.999) * self.distance
                )

            self.sb.update_transform(context)

        return {"RUNNING_MODAL"}
