import webbrowser

import bpy
from bpy.types import AddonPreferences

import rna_keymap_ui

from .props import sectionbox_settings
from .icons import icon_id
from .ui import unregister_ui, register_ui
from .utils.preferences import get_def_material_path, remove_old_def_materials
from .keymaps import keys
from . import addon_id

bl_info = {
    "name": "Section Box",
    "description": "Visualization, Sections, Elevations, and more vfxMed",
    "blender": (3, 4, 0),
    "version": (2, 1, 2),
    "category": "3D View",
    "author": "Ruben Messerschmidt",
    "location": "View3d > Sidebar > Section Box",
    "doc_url": "https://section-box.readthedocs.io/en/latest/",
    "version_url": "https://raw.githubusercontent.com/rubenmesserschmidt/Section-Box_docs/main/version",
}


class Email(bpy.types.Operator):
    bl_idname = "sectionbox.email"
    bl_label = "Send an Email"
    bl_options = {"INTERNAL"}
    bl_description = "Send an Email to the developer"

    def execute(self, context):
        addon_name = bl_info["name"]
        addon_version = (
            f"{bl_info['version'][0]}.{bl_info['version'][1]}.{bl_info['version'][2]}"
        )
        blender_version = (
            f"{bpy.app.version[0]}.{bpy.app.version[1]}.{bpy.app.version[2]}"
        )
        os = bpy.app.build_platform.decode("utf-8")
        mail_to = (
            f"mailto:support@rubenmesserschmidt.com?subject=%5B{addon_name}%5D%20&body="
            f"Addon%20Version%3A%20{addon_version}%0A"
            f"Blender%20Version%3A%20{blender_version}%0A"
            f"OS%3A%20{os}%0A%0A"
            f"Please%20replace%20this%20text%20with%20your%20question%20or%20issue.%20For%20any%20bugs%2C%20please%20give%20me%20an%20instruction%20on%20how%20I%20can%20reproduce%20it%20and%20share%20the%20file%20that's%20not%20working%2C%20if%20possible.%20Thank%20you!%0A"
        )
        webbrowser.open(mail_to)
        return {"FINISHED"}


def update_merge_panels(self, context):
    unregister_ui()
    register_ui()


class AddonPreferences(AddonPreferences):
    bl_idname = __package__

    # common settings
    update_check: bpy.props.BoolProperty(
        name="Check for updates on startup",
        default=False,
        description="Automatically check for updates every time you start Blender",
    )
    ignore_update: bpy.props.FloatVectorProperty(
        name="Ignore Update",
        default=(0, 0, 0),
    )
    merge_panels: bpy.props.BoolProperty(
        name="Merge Panels",
        default=False,
        description='Merge the Instant Clean panel into the "Ruben\'s Addons" tab',
        update=lambda s, c: update_merge_panels(s, c),
    )

    # specific addon settings
    show_menu: bpy.props.BoolProperty(
        name="Show Menu",
        default=True,
        description="Show the create section box operator in the object context menu",
    )

    select_radius: bpy.props.IntProperty(
        name="Select Radius",
        default=30,
        min=10,
        description="Radius of the hitbox of the section box handles",
    )

    lazy_load: bpy.props.BoolProperty(
        name="Lazy Load",
        default=True,
        description="Update changes only after committing a operation to increase performance while editing",
    )

    margin: bpy.props.FloatProperty(
        name="Margin Size",
        default=0.01,
        min=0.0,
        description="Additional size for the section box to prevent cutting the objects in the initial transformation",
    )

    empty_size: bpy.props.FloatProperty(
        name="Empty Size",
        default=1,
        min=0.01,
        description="Size of the section box when no objects are selected while creating",
    )

    handle_color: bpy.props.FloatVectorProperty(
        name="Handle Color",
        size=4,
        default=(0, 0, 0, 0.5),
        min=0,
        max=1,
        subtype="COLOR",
        description="Color of the section box handles",
    )

    handle_color_active: bpy.props.FloatVectorProperty(
        name="Handle Color Active",
        size=4,
        default=(0, 0, 0, 1),
        min=0,
        max=1,
        subtype="COLOR",
        description="Color of the active section box handle",
    )

    handle_hover_scale: bpy.props.FloatProperty(
        name="Handle Hover Scale",
        default=1.2,
        min=0,
        description="Scale of the section box handles when hovering over them",
    )

    # box default settings
    def_settings: bpy.props.PointerProperty(
        type=sectionbox_settings.SECTIONBOX_PG_SectionBoxSettings
    )
    def_method: bpy.props.EnumProperty(
        name="Method",
        items=[
            (
                "MODE_BASED",
                "Mode Based",
                "Use the suitable section display method for the current viewport shading mode",
                "VIEW3D",
                0,
            ),
            ("BOOLEAN", "Boolean", "", "MOD_BOOLEAN", 1),
            ("SHADER", "Shader", "", "SHADING_TEXTURE", 2),
        ],
        description="Default section display method. Choose 'Mode Based' to use the suitable method for the current viewport shading mode",
    )
    def_material: bpy.props.StringProperty(
        default="From Mesh",
    )

    spacing_right = 4.8
    sub1_spacing = 2.5
    split_factor = 0.5

    def export_def_material(self):
        if self.def_material.endswith("[Section Box]"):
            return

        path = get_def_material_path(addon_id)
        if path.exists():
            path.unlink()

        remove_old_def_materials()
        if self.def_material == "From Mesh":
            return

        mat = bpy.data.materials.get(self.def_material)
        if not mat:
            return

        def_mat = mat.copy()
        name = f"{mat.name} [Section Box]"
        def_mat.name = name
        self.def_material = name
        bpy.data.libraries.write(
            filepath=str(path),
            datablocks={bpy.data.materials[self.def_material]},
            fake_user=True,
        )

    def import_def_material(self):
        path = get_def_material_path(addon_id)
        if path.exists():
            with bpy.data.libraries.load(str(path)) as (data_from, data_to):
                def_mat = data_from.materials[0]
                if not bpy.data.materials.get(def_mat):
                    data_to.materials = data_from.materials

    def draw(self, context):
        layout = self.layout

        # general settings
        col = layout.column()
        row = col.row()
        row.enabled = False
        row.label(text="General Settings")
        split = col.split()
        sub = split.column()
        sub.label(text="Lazy Load")
        sub.label(text="Method")
        sub.label(text="Margin")
        sub.label(text="Empty Size")
        sub = split.column()
        sub.prop(self, "lazy_load", text="")
        sub.prop(self, "def_method", text="")
        sub.prop(self, "margin", text="")
        sub.prop(self, "empty_size", text="")
        col.separator(factor=3)

        # UI settings
        col = layout.column()
        row = col.row()
        row.enabled = False
        row.label(text="UI Settings")

        split = col.split(factor=self.split_factor)
        sub = split.column()
        sub.label(text="Show Menu")
        sub.label(text="Select Radius")
        sub.label(text="Handle Color")
        sub.label(text="Handle Color (Active)")
        sub.label(text="Handle Hover Scale")
        sub = split.column()
        sub.prop(self, "show_menu", text="")
        sub.prop(self, "select_radius", text="")
        sub.prop(self, "handle_color", text="")
        sub.prop(self, "handle_color_active", text="")
        sub.prop(self, "handle_hover_scale", text="")
        col.separator(factor=3)

        # keymap settings
        col = layout.column()
        row = col.row()
        row.enabled = False
        row.label(text="Keymap")
        col.separator()
        wm = bpy.context.window_manager
        kc = wm.keyconfigs.user
        old_id_l = []

        for key in keys:
            km = kc.keymaps.get(key.get("keymap"))
            if km:
                kmi = km.keymap_items.get(key.get("idname"))
                if kmi:
                    for km_con in kc.keymaps:
                        if km.name == km_con.name:
                            km = km_con
                            break

                    for kmi_con in km.keymap_items:
                        if kmi.idname == kmi_con.idname:
                            if not kmi_con.id in old_id_l:
                                kmi = kmi_con
                                old_id_l.append(kmi_con.id)
                                break

                    try:
                        col.context_pointer_set("keymap", km)
                        rna_keymap_ui.draw_kmi([], kc, km, kmi, col, 0)
                        col.separator()

                    except:
                        pass

        # merge panels
        layout.separator()
        row = layout.row()
        row.enabled = False
        row.label(text="Ruben's Addons")

        split = layout.split()
        col = split.column()
        col.label(text="Merge Panels")
        col = split.column()
        col.prop(self, "merge_panels", text="")

        # update
        layout.separator()
        row = layout.row()
        row.enabled = False
        row.label(text="Update")
        split = layout.split()
        col = split.column()
        col.label(text="Check for updates")
        col.label(text="Auto check on startup")
        col = split.column()
        col.operator(
            "sectionbox.update_check", text="Check for updates", icon="FILE_REFRESH"
        ).manual = True
        col.prop(self, "update_check", text="")

        # support
        layout.separator()
        row = layout.row()
        row.enabled = False
        row.label(text="Support")
        split = layout.split()
        col = split.column()
        col.label(text="Join Discord")
        col.label(text="Send an E-Mail")
        col = split.column()
        col.operator(
            f"sectionbox.support", text="Discord", icon_value=icon_id("discord.png")
        ).url = "https://discord.gg/TYFq6VKXCN"
        col.operator(f"sectionbox.support", text="E-Mail", icon="WORDWRAP_ON").url = (
            "mailto:support@rubenmesserschmidt.com"
        )
