import math
import time
from pathlib import Path
from mathutils import Vector, Matrix
from itertools import chain

import bpy
from bpy.types import PropertyGroup
import bmesh

from .section_plane import SECTIONBOX_PG_SectionPlane
from .included_object import SECTIONBOX_PG_IncludedObject
from .view import SECTIONBOX_PG_View
from .sectionbox_settings import (
    SECTIONBOX_PG_SectionBoxSettings,
    update_cross_section,
    update_elevation,
    update_material,
)
from ..utils.aabb import AABB
from ..utils.collection import get_child_collection, close_collection
from ..utils.objects import get_evaluated_object, is_geo
from ..utils.mouse import mouse_hit, location_3d_to_2d
from ..utils.settings import store_settings, restore_settings
from ..utils.collection import link_to_collection, unlink_from_collection
from ..utils.preferences import get_prefs
from ..utils.properties import prop_filter
from .. import addon_id


class SECTIONBOX_PG_SectionBox(SECTIONBOX_PG_SectionBoxSettings, PropertyGroup):
    id: bpy.props.IntProperty()
    entry_name: bpy.props.StringProperty()
    object: bpy.props.PointerProperty(type=bpy.types.Object)
    planes: bpy.props.CollectionProperty(type=SECTIONBOX_PG_SectionPlane)
    active_plane: bpy.props.IntProperty()
    cutter: bpy.props.PointerProperty(type=bpy.types.Object)
    views: bpy.props.CollectionProperty(type=SECTIONBOX_PG_View)
    active_view: bpy.props.IntProperty(default=-1)
    collection: bpy.props.PointerProperty(type=bpy.types.Collection)
    included_objects: bpy.props.CollectionProperty(type=SECTIONBOX_PG_IncludedObject)
    sb_collection: bpy.props.PointerProperty(type=bpy.types.Collection)
    tmp_objects_collection: bpy.props.PointerProperty(type=bpy.types.Collection)
    drawings_collection: bpy.props.PointerProperty(type=bpy.types.Collection)
    aabb: bpy.props.FloatVectorProperty(size=24)

    # settings
    pre_settings: bpy.props.PointerProperty(type=SECTIONBOX_PG_SectionBoxSettings)

    def set_update(self, update):
        self.update_ = update
        for plane in self.planes:
            plane.update_ = update

    def get_drawing_prop(self, drawing, prop, link):
        if link:
            return getattr(self, f"cross_section_{prop}")
        else:
            return getattr(self, f"{drawing}_{prop}")

    def is_active(self, context):
        return (
            self.object is context.view_layer.objects.active
            and self.object.select_get()
        )

    def iter_affected_objects(self):
        for io in self.included_objects:
            yield io.get_affected_object()

    def set_bounding_box(self, context, margin):
        # create aabb
        prefs = get_prefs()
        aabb = AABB()
        aabb.from_objects(
            context,
            list(self.iter_affected_objects()),
            margin=margin,
            empty_size=prefs.empty_size,
        )
        aabb.to_mesh(self.object.data)

        # apply bounding box to section box object
        self.object.location = aabb.center()
        self.aabb = list(chain.from_iterable(aabb.get_vertices()))

        for vert in self.object.data.vertices:
            vert.co -= self.object.location

        return aabb

    def init(self, context, objects):
        prefs = get_prefs()

        # create collection hierarchy
        scene_col = context.scene.sectionbox.collection
        self.sb_collection = get_child_collection(
            context, scene_col, f"{scene_col.name}.{self.name}"
        )
        self.sb_collection.hide_render = True

        self.tmp_objects_collection = get_child_collection(
            context, self.sb_collection, f"{self.sb_collection.name}.tmp_objs"
        )
        self.tmp_objects_collection.hide_select = True

        self.drawings_collection = get_child_collection(
            context, self.sb_collection, f"{self.sb_collection.name}.drawings"
        )

        # close collection hierarchy
        close_collection(context, self.sb_collection)
        close_collection(context, context.scene.sectionbox.collection)

        # prepare section box object
        self.sb_collection.objects.link(self.object)
        self.object.display_type = "WIRE"
        self.object.hide_render = True

        # create transparent material
        sb_material_name = "sectionbox_material"
        sb_material = bpy.data.materials.get(
            sb_material_name
        ) or bpy.data.materials.new(sb_material_name)
        sb_material.use_nodes = True
        nodes = sb_material.node_tree.nodes
        bsdf_node = nodes.new("ShaderNodeBsdfTransparent")

        # Connect the Principled BSDF node to the material output
        sb_material.node_tree.links.new(
            bsdf_node.outputs["BSDF"], nodes["Material Output"].inputs["Surface"]
        )

        self.object.data.materials.append(sb_material)

        # load settings
        prefs = get_prefs()
        for prop in [
            p
            for p in filter(
                lambda a: prop_filter(prefs.def_settings, a), dir(prefs.def_settings)
            )
            if p not in {"method", "material"}
        ]:
            setattr(self, prop, getattr(prefs.def_settings, prop))

        mat = prefs.def_material
        if mat != "From Mesh":
            if not bpy.data.materials.get(mat):
                prefs.import_def_material()

            self.material = mat

        if self.collection:
            self.method = "BOOLEAN"
        elif context.space_data:
            if prefs.def_method == "MODE_BASED":
                self.method = (
                    "SHADER"
                    if context.space_data.shading.type in {"MATERIAL"}
                    else "BOOLEAN"
                )
            else:
                self.method = prefs.def_method

        objs = [o for o in objects]
        # evaluate included objects
        for obj in objs:
            self.add_included_object(context, obj)

        # create aabb
        aabb = self.set_bounding_box(context, prefs.margin)

        # if empty section box
        if not self.included_objects:
            self.object.location = context.scene.cursor.location
            objects = aabb.get_intersecting_objects(context, self.object.location)
            for obj in objects:
                self.add_included_object(context, obj)

        # create cutter object
        self.cutter = self.object.copy()
        self.cutter.data = self.object.data.copy()
        self.sb_collection.objects.link(self.cutter)
        self.cutter.location = self.object.location
        cutter_name = f"{self.name}.cutter"
        self.cutter.name = cutter_name
        self.cutter.data.name = cutter_name
        self.cutter.hide_set(True)
        self.cutter.hide_select = True
        self.cutter.hide_render = True
        self.cutter.hide_viewport = True

        # mark freestyle faces
        for polygon in self.cutter.data.polygons:
            polygon.use_freestyle_mark = True

        # create section planes
        for i in range(6):
            plane = self.planes.add()
            plane.id = i
            plane.sb_name = self.name
            plane.init(context)

        for io in self.included_objects:
            self.setup_included_object(context, io)

        self.set_update(True)

        # force boolean update
        bpy.app.timers.register(
            lambda: setattr(
                self.cutter,
                "location",
                self.cutter.location + Vector((0, 0, 1e-24)),
            ),
            first_interval=0.01,
        )

    def add_object(self, context: bpy.types.Context, object: bpy.types.Object):
        included_object = self.add_included_object(context, object)
        if not included_object:
            return

        self.update_aabb(context)
        self.setup_included_object(context, included_object)

        # update intersecting
        self.update_planes_intersecting(context)

    def add_included_object(self, context, object):
        # check if object is a section box object
        if object in context.scene.sectionbox.collection.all_objects.values():
            return

        tmp_object = get_evaluated_object(context, object)
        if not (tmp_object or is_geo(object)):
            return

        item = self.included_objects.add()
        item.object = object
        item.set_tmp_object(tmp_object, self.tmp_objects_collection)
        return item

    def setup_included_object(
        self, context: bpy.types.Context, included_object: SECTIONBOX_PG_IncludedObject
    ):
        object = included_object.get_affected_object()

        # get section material
        section_material = bpy.data.materials.get(f"{self.name}.material")
        if not section_material:
            section_material = bpy.data.materials.new(f"{self.name}.material")
            section_material.use_nodes = True
            section_material.blend_method = "CLIP"
            node_tree = section_material.node_tree
            node_tree.nodes.clear()
            diffuse = node_tree.nodes.new("ShaderNodeBsdfDiffuse")
            output = node_tree.nodes.new("ShaderNodeOutputMaterial")
            node_tree.links.new(diffuse.outputs[0], output.inputs[0])

        # add section material if no material is assigned
        if not object.data.materials:
            object.data.materials.append(section_material)

        # set flag to exclude non affected objects with same material from section
        object[self.name] = 1

        # add section node group to materials
        for mat in object.data.materials:
            if not mat:
                continue
            if not mat.get(self.name):
                mat[self.name] = 1
            else:
                mat[self.name] += 1

            if not mat.node_tree:
                mat.use_nodes = True

            if mat.blend_method != "CLIP":
                mat["prev_blend_method"] = mat.blend_method
                mat.blend_method = "CLIP"

            node_tree = mat.node_tree
            # check if material already prepared
            if node_tree.nodes.get(self.name):
                continue

            # section box node group
            node_group = node_tree.nodes.new("ShaderNodeGroup")
            node_group.node_tree = bpy.data.node_groups["section_node_group"]
            node_group.name = self.name
            node_group.label = self.name
            node_group.mute = self.method != "SHADER"

            output = [n for n in node_tree.nodes if n.type == "OUTPUT_MATERIAL"]
            if output:
                output = [o for o in output if o.is_active_output][0]

            if not output:
                output = node_tree.nodes.new("ShaderNodeOutputMaterial")

            try:
                input_socket = output.inputs[0].links[0].from_socket
            except:
                pass
            else:
                node_tree.links.new(input_socket, node_group.inputs[0])
            finally:
                node_tree.links.new(node_group.outputs[0], output.inputs[0])

            # attribute node
            attribute = node_tree.nodes.new("ShaderNodeAttribute")
            attribute.name = f"{self.name}.attribute"
            attribute.attribute_type = "OBJECT"
            attribute.attribute_name = self.name

            node_tree.links.new(attribute.outputs[2], node_group.inputs["ID"])

            # texture coordinate nodes
            for plane in self.planes:
                texture_coordinate = node_tree.nodes.new("ShaderNodeTexCoord")
                texture_coordinate.object = plane.object
                texture_coordinate.name = f"{self.name}.plane_{plane.id}"
                texture_coordinate.hide = True
                node_tree.links.new(
                    texture_coordinate.outputs[3], node_group.inputs[str(plane.id)]
                )

        # add boolean modifier
        mod = object.modifiers.new(f"{self.name}.boolean", type="BOOLEAN")
        mod.object = self.cutter
        mod.operation = "DIFFERENCE" if self.invert else "INTERSECT"
        mod.solver = "EXACT"
        mod.material_mode = "TRANSFER"
        mod.use_self = self.boolean_use_self
        mod.use_hole_tolerant = self.boolean_use_hole_tolerant
        mod.show_viewport = self.enable and self.method != "SHADER"
        mod.show_render = (
            self.enable and not self.hide_render
        ) and self.method != "SHADER"

    def get_included_object(self, object):
        for included_object in self.included_objects:
            if object in (included_object.object, included_object.tmp_object):
                return included_object

    def remove_object(self, context, object, recalculate=True):
        res = [
            (i, io) for i, io in enumerate(self.included_objects) if object == io.object
        ]
        if not res:
            return

        id, included_object = res[0]
        included_object.delete(context, self)
        self.included_objects.remove(id)

        # update intersecting
        if recalculate:
            self.update_aabb(context)
            self.update_planes_intersecting(context)

    def update_aabb(self, context):
        aabb = AABB()
        aabb.from_objects(context, self.iter_affected_objects())
        self.aabb = list(chain.from_iterable(aabb.get_vertices()))

    def update_planes_intersecting(self, context):
        for plane in self.planes:
            plane.update_intersecting(context)

    def delete_data(self, context):
        for plane in self.planes:
            plane.delete(context)

        if self.cutter:
            bpy.data.meshes.remove(self.cutter.data)
        section_mat = bpy.data.materials.get(f"{self.name}.material")
        if section_mat:
            bpy.data.materials.remove(section_mat)
        if self.tmp_objects_collection:
            bpy.data.collections.remove(self.tmp_objects_collection)
        if self.drawings_collection:
            bpy.data.collections.remove(self.drawings_collection)
        if self.sb_collection:
            bpy.data.collections.remove(self.sb_collection)

        # this was the last sectionbox
        if len(context.scene.sectionbox.list) == 1:
            context.window_manager.sectionbox.running_modals = False

            cs_lineset = context.view_layer.freestyle_settings.linesets.get(
                "sectionbox.cross_section"
            )
            if cs_lineset:
                context.view_layer.freestyle_settings.linesets.remove(cs_lineset)
            ele_lineset = context.view_layer.freestyle_settings.linesets.get(
                "sectionbox.elevation"
            )
            if ele_lineset:
                context.view_layer.freestyle_settings.linesets.remove(ele_lineset)
            freestyle_flag_mat = bpy.data.materials.get("sectionbox.freestyle_flag")
            if freestyle_flag_mat:
                bpy.data.materials.remove(freestyle_flag_mat)
            ex_col = context.scene.sectionbox.exclude_render_collection
            if ex_col:
                bpy.data.collections.remove(ex_col)
            in_col = context.scene.sectionbox.include_render_collection
            if in_col:
                bpy.data.collections.remove(in_col)
            bpy.data.collections.remove(context.scene.sectionbox.collection)
            sb_node_group = bpy.data.node_groups.get("section_node_group")
            if sb_node_group:
                bpy.data.node_groups.remove(sb_node_group)

            # remove draw handler
            dh = bpy.app.driver_namespace.get("sectionbox.draw_handles")
            if dh:
                dh.remove_handler()
                del bpy.app.driver_namespace["sectionbox.draw_handles"]

            # remove transparent material
            sb_material = bpy.data.materials.get("sectionbox_material")
            if sb_material:
                bpy.data.materials.remove(sb_material)

        # delete section box from list
        context.scene.sectionbox.list.remove(context.scene.sectionbox.list_index)
        context.scene.sectionbox.list_index -= (
            1 if context.scene.sectionbox.list_index > 0 else 0
        )

        # delete section box object
        bpy.data.meshes.remove(self.object.data)

    def delete(self, context):
        objects = [io.object for io in self.included_objects]
        for obj in objects:
            self.remove_object(context, obj, recalculate=False)

        self.delete_data(context)

    def add_view(self, name):
        tmp_name = name
        i = 1
        while any(tmp_name == v.name for v in self.views):
            tmp_name = "{name}.{i}".format(name=name, i=("00" + str(i))[-3:])
            i += 1

        name = tmp_name
        view = self.views.add()
        view.name = name
        view.init(self)
        self.active_view = len(self.views) - 1
        return view

    def remove_view(self, view_id):
        self.views.remove(view_id)
        self.active_view = -1

    def set_view(self, context, view_id):
        self.active_view = view_id
        view = self.views[view_id]
        for prop in filter(
            lambda a: prop_filter(view.sectionbox_settings, a),
            dir(view.sectionbox_settings),
        ):
            try:
                setattr(self, prop, getattr(view.sectionbox_settings, prop))
            except:
                continue

        for plane_settings, plane in zip(view.plane_settings, self.planes):
            plane._update = False
            for prop in filter(
                lambda a: prop_filter(plane_settings, a), dir(plane_settings)
            ):
                try:
                    setattr(plane, prop, getattr(plane_settings, prop))
                except:
                    continue

            plane._update = True

        pre_active_plane = self.active_plane
        for plane in self.planes:
            self.active_plane = plane.id
            self.update_transform(context)

        self.active_plane = pre_active_plane
        self.update_cutter(context)

    def reload(self, context):
        if not self.collection:
            return

        for obj in self.collection.all_objects:
            if not self.get_included_object(obj):
                self.add_object(context, obj)

    def reset(self, context):
        prefs = get_prefs()

        # disable boolean modifiers
        objs = [o for o in self.iter_affected_objects()]
        for obj in objs:
            mod = obj.modifiers.get(f"{self.name}.boolean")
            if mod:
                mod.show_viewport = False

        # calculate bounding box
        aabb = AABB()
        aabb.from_objects(
            context, objs, margin=prefs.margin, empty_size=prefs.empty_size
        )

        # reset object transform
        self.object.rotation_euler = Vector()
        self.object.scale = Vector((1, 1, 1))
        self.object.location = aabb.center()

        # reset mesh transform
        for vert, co in zip(self.object.data.vertices, aabb.get_vertices()):
            vert.co = co - self.object.location

        # reenable boolean modifiers
        for obj in self.iter_affected_objects():
            mod = obj.modifiers.get(f"{self.name}.boolean")
            if mod:
                mod.show_viewport = True

        # reset planes transform
        for plane in self.planes:
            plane.auto_update = False
            plane.transform = 0
            plane.edge_transform = [0, 0, 0, 0]
            plane.auto_update = True

        # reset cutter transform
        self.update_cutter(context)

    def update_transform(self, context):
        plane_axis = {
            0: 2,
            1: 2,
            2: 1,
            3: 1,
            4: 0,
            5: 0,
        }

        # update box transform
        face = self.object.data.polygons[self.active_plane]
        plane = self.planes[self.active_plane]
        sign = -1 if self.active_plane % 2 else 1
        axis = plane_axis[self.active_plane]
        for i in range(4):
            vert = face.vertices[i]
            k = i - 1
            self.object.data.vertices[vert].co[axis] = (
                plane.initial_location[axis]
                + sign * plane.transform
                + sign * plane.edge_transform[i]
                + sign * plane.edge_transform[k]
            )

        # update planes transform
        for plane in self.planes:
            plane.update_transform(context)

    def update_cutter(self, context):
        # update cutter mesh
        for c_vert, o_vert in zip(self.cutter.data.vertices, self.object.data.vertices):
            c_vert.co = o_vert.co

        # mark freestyle faces
        for polygon in self.cutter.data.polygons:
            polygon.use_freestyle_mark = True

        update_material(self, context)

        plane = self.planes[self.active_plane]
        plane.update_intersecting(context)

    def iter_affected_materials(self):
        for obj in self.iter_affected_objects():
            if not obj:
                continue

            for m in obj.data.materials:
                if not m:
                    continue

                yield m

    def hit_plane(self, context, mouse_co, select=True):
        index = mouse_hit(
            mouse_co,
            [
                location_3d_to_2d(context, self.object.matrix_world @ p.object.location)
                for p in self.planes
            ],
            context.preferences.addons[addon_id].preferences.select_radius,
        )
        if index is None:
            return None

        if select:
            self.active_plane = index

        return self.planes[index]

    def opposite_plane(self, plane):
        opposite = plane.id - 1 if plane.id % 2 else plane.id + 1
        return self.planes[opposite]

    def opposite_edge(self, plane, edge):
        opposite_plane = self.opposite_plane(plane)
        opposite_face = self.object.data.polygons[opposite_plane.id]
        face = self.object.data.polygons[plane.id]
        adj_face = [
            f
            for f in self.object.data.polygons
            if edge.index
            in [self.object.data.edge_keys.index(ek) for ek in f.edge_keys]
            and f.index != face.index
        ][0]
        for edge in [
            self.object.data.edges[self.object.data.edge_keys.index(ek)]
            for ek in opposite_face.edge_keys
        ]:
            if edge.index in [
                self.object.data.edge_keys.index(ek) for ek in adj_face.edge_keys
            ]:
                return edge

    def prepare_for_render(self, context):
        # hide objects
        for obj in [
            o
            for o in context.scene.collection.all_objects
            if o not in self.iter_affected_objects()
        ]:
            if obj.hide_render:
                obj["sectionbox.prev_hide_render"] = True
            else:
                obj.hide_render = True

        # hide collections
        for col in [
            c
            for c in context.scene.collection.children
            if c not in {context.scene.sectionbox.collection, self.collection}
        ]:
            col["sectionbox.prev_hide_render"] = col.hide_render
            col["sectionbox.prev_hide_viewport"] = col.hide_viewport
            col.hide_render = True
            col.hide_viewport = True

        # link objects
        for obj in self.iter_affected_objects():
            context.scene.sectionbox.collection.objects.link(obj)

        # set method to boolean if not already
        self.prev_method = None
        if self.method != "BOOLEAN":
            self.prev_method = self.method
            self.method = "BOOLEAN"

    def restore_after_render(self, context):
        # unlink objects
        for obj in self.iter_affected_objects():
            context.scene.sectionbox.collection.objects.unlink(obj)

        # unhide objects
        for obj in [
            o
            for o in context.scene.collection.all_objects
            if o not in self.iter_affected_objects()
        ]:
            if obj.get("sectionbox.prev_hide_render", None):
                del obj["sectionbox.prev_hide_render"]
            else:
                obj.hide_render = False

        # unhide collections
        for col in [
            c
            for c in context.scene.collection.children
            if c not in {context.scene.sectionbox.collection, self.collection}
        ]:
            col.hide_render = col.get("sectionbox.prev_hide_render", False)
            col.hide_viewport = col.get("sectionbox.prev_hide_viewport", False)
            del col["sectionbox.prev_hide_render"]
            del col["sectionbox.prev_hide_viewport"]

        # restore method if was not boolean
        if self.prev_method:
            self.method = self.prev_method
            self.prev_method = None

    def apply_scale(self, context):
        scale_matrix = Matrix.Scale(1, 3)
        scale_matrix[0][0] = self.object.scale.x
        scale_matrix[1][1] = self.object.scale.y
        scale_matrix[2][2] = self.object.scale.z
        for v in self.object.data.vertices:
            v.co = scale_matrix @ v.co

        self.object.scale = Vector((1, 1, 1))

    def recover_scale(self, context, scale):
        scale_matrix = Matrix.Scale(1, 3)
        scale_matrix[0][0] = 1 / scale.x
        scale_matrix[1][1] = 1 / scale.y
        scale_matrix[2][2] = 1 / scale.z
        for v in self.object.data.vertices:
            v.co = scale_matrix @ v.co

        self.object.scale = scale

    def cross_section(self, context, plane):
        name = f"cross_section.{self.name}.{plane.id}"

        # TODO: replace with context override everywhere
        # prepare context
        pre_active = context.view_layer.objects.active
        pre_selected = context.selected_objects[:]
        bpy.ops.object.select_all(action="DESELECT")

        # apply scale
        prev_scale = self.object.scale.copy()
        self.apply_scale(context)

        # join objects
        bm = bmesh.new()
        depsgraph = context.evaluated_depsgraph_get()
        for obj in self.iter_affected_objects():
            tmp_obj = obj.evaluated_get(depsgraph)
            for vert in tmp_obj.data.vertices:
                vert.co = obj.matrix_world @ vert.co

            bm.from_mesh(tmp_obj.data)

        # bisect at plane
        face = self.object.data.polygons[plane.id]
        mw = self.object.matrix_world
        normal = (mw @ face.normal) - (mw @ Vector((0, 0, 0)))
        bmesh.ops.bisect_plane(
            bm,
            geom=list(bm.verts) + list(bm.edges),
            dist=0.0001,
            plane_co=mw @ face.center,
            plane_no=normal,
            clear_outer=True,
            clear_inner=True,
        )

        if not self.cross_section_entire:
            for other_plane in [p for p in self.planes if p is not plane]:
                other_face = self.object.data.polygons[other_plane.id]
                other_normal = (mw @ other_face.normal) - (mw @ Vector((0, 0, 0)))
                bmesh.ops.bisect_plane(
                    bm,
                    geom=list(bm.verts) + list(bm.edges),
                    dist=0.0001,
                    plane_co=mw @ other_face.center,
                    plane_no=other_normal,
                    clear_outer=True,
                    clear_inner=False,
                )

        if not bm.edges:
            bm.free()
            return False

        # remove face holding edges
        face_edges = []
        for edge in bm.edges:
            if len(edge.link_faces) > 1:
                face_edges.append(edge)

        for edge in face_edges:
            bm.edges.remove(edge)

        # remove faces
        for face in bm.faces:
            bm.faces.remove(face)

        # remove straight verts
        straight_verts = [
            v
            for v in bm.verts
            if len(v.link_edges) == 2
            and -1e-3
            < (v.link_edges[0].other_vert(v).co - v.co).angle(
                v.link_edges[1].other_vert(v).co - v.co
            )
            - math.pi
            < 1e-3
        ]
        bmesh.ops.dissolve_verts(bm, verts=straight_verts)

        # return True

        mesh = bpy.data.meshes.new(name)
        bm.to_mesh(mesh)
        bm.free()
        cs_obj = bpy.data.objects.new(name, mesh)
        plane.cross_section = cs_obj
        self.drawings_collection.objects.link(cs_obj)

        # return True

        # create gpencil object
        gpencil_mesh_obj = cs_obj.copy()
        gpencil_mesh_obj.data = cs_obj.data.copy()
        self.drawings_collection.objects.link(gpencil_mesh_obj)
        context.view_layer.objects.active = gpencil_mesh_obj
        gpencil_mesh_obj.select_set(True)
        bpy.ops.object.transform_apply(
            location=False, scale=True, rotation=False, properties=False
        )
        bpy.ops.object.convert(
            target="GPENCIL", keep_original=True, thickness=3, offset=0
        )
        cs_obj.hide_render = True
        cs_obj.hide_viewport = True
        cs_obj.hide_select = True

        # remove duplicated mesh
        bpy.data.meshes.remove(gpencil_mesh_obj.data)

        # transform cs_obj into xy plane for export
        for vert in cs_obj.data.vertices:
            vert.co = plane.object.matrix_world.inverted() @ vert.co

        # recover scale
        self.recover_scale(context, prev_scale)

        # prepare gpencil object
        cs_gp = context.view_layer.objects.active
        prev_col = [
            c
            for c in list((context.scene.collection,))
            + list(context.scene.collection.children)
            if cs_gp in c.objects.values()
        ]
        if prev_col:
            try:
                prev_col[0].objects.unlink(cs_gp)
            except Exception as e:
                print(f"Could not unlink GPencil object: {e}")

        cs_gp.data.layers.active.tint_factor = 1
        cs_gp.hide_select = True
        cs_gp.name = f"cross_section.{self.name}.{plane.id}.gpencil"
        cs_gp.data.name = cs_gp.name

        self.drawings_collection.objects.link(cs_gp)
        plane.cross_section_gpencil = cs_gp
        update_cross_section(self, context, "dst", (plane,))
        update_cross_section(self, context, "thickness", (plane,))
        update_cross_section(self, context, "color", (plane,))

        context.view_layer.objects.active = pre_active
        [o.select_set(True) for o in pre_selected]

        if bpy.app.version >= (4, 1, 0):
            self.update_cutter(context)

        return True

    def elevation(self, context, plane):
        lineset = context.view_layer.freestyle_settings.linesets.get(
            "sectionbox.elevation"
        ) or context.view_layer.freestyle_settings.linesets.new("sectionbox.elevation")

        lineset.select_silhouette = False
        lineset.select_border = True
        lineset.select_crease = True
        lineset.select_material_boundary = True
        lineset.select_external_contour = True
        lineset.show_render = False
        lineset.select_by_face_marks = True
        lineset.face_mark_negation = "EXCLUSIVE"

        # set camera clip end
        farthest_dst = 1e-6
        cam = plane.camera
        if self.invert:
            for obj in self.iter_affected_objects():
                for vert in obj.data.vertices:
                    dst = (obj.matrix_world @ vert.co - cam.location).length
                    if dst > farthest_dst:
                        farthest_dst = dst
        else:
            cam_loc = cam.matrix_world @ cam.location
            for vert in self.object.data.vertices:
                dst = (self.object.matrix_world @ vert.co - cam_loc).length
                if dst > farthest_dst:
                    farthest_dst = dst

        cam.data.clip_end = farthest_dst

        # create drawing
        result = self.create_drawing(context, plane, lineset)
        if not result:
            return False

        ele_obj, ele_gp = result
        ele_gp.data.layers.active.tint_factor = 1
        ele_obj.name = f"elevation.{self.name}.{plane.id}"
        ele_gp.name = f"elevation.{self.name}.{plane.id}.gpencil"

        plane.elevation = ele_obj
        plane.elevation_gpencil = ele_gp
        update_elevation(self, context, "dst", (plane,))
        update_elevation(self, context, "thickness", (plane,))
        update_elevation(self, context, "color", (plane,))
        return True

    def create_drawing(self, context, plane, lineset):
        self.prepare_for_render(context)

        # set camera active
        store_settings(context.scene, properties={"camera"})
        plane = self.planes[self.active_plane]
        context.scene.camera = plane.camera

        # set render settings
        store_settings(context.scene.render)
        context.scene.render.engine = "BLENDER_WORKBENCH"
        context.scene.render.use_freestyle = True
        context.scene.render.line_thickness = 1
        context.scene.render.film_transparent = True
        context.scene.svg_export.use_svg_export = True
        context.scene.render.resolution_x = 1000
        context.scene.render.resolution_y = 1000
        context.scene.render.resolution_percentage = 100

        # save prev lineset settings
        for lineset in context.view_layer.freestyle_settings.linesets:
            store_settings(lineset, properties={"show_render"})

        # disable all linesets
        for lineset in context.view_layer.freestyle_settings.linesets:
            lineset.show_render = False

        # enable lineset
        lineset.show_render = True

        # freestyle settings
        store_settings(
            context.view_layer.freestyle_settings,
            properties={"crease_angle", "use_culling"},
        )
        context.view_layer.freestyle_settings.crease_angle = self.elevation_crease_angle
        context.view_layer.freestyle_settings.use_culling = False

        # set svg output path
        path = Path(bpy.utils.user_resource("DATAFILES", path=addon_id, create=True))
        for file in path.glob("drawing*.svg"):
            file.unlink()

        for col in [
            c
            for c in bpy.data.collections
            if c.name.startswith("drawing") and c.name.endswith(".svg")
        ]:
            bpy.data.collections.remove(col)

        context.scene.render.filepath = str(path / "drawing.svg")
        bpy.ops.render.render(use_viewport=True)

        while not [f for f in path.glob("drawing*.svg")]:
            time.sleep(0.1)

        # restore prev lineset settings
        for lineset in context.view_layer.freestyle_settings.linesets:
            restore_settings(lineset)

        # restore prev camera
        restore_settings(context.scene, properties={"camera"})

        # restore prev render settings
        restore_settings(context.scene.render)

        # restore prev freestyle settings
        restore_settings(context.view_layer.freestyle_settings)

        # restore object visibility
        self.restore_after_render(context)

        # import drawing
        for file in path.glob("drawing*.svg"):
            bpy.ops.import_curve.svg(filepath=str(file))
            break

        drawing_col = [
            c
            for c in bpy.data.collections
            if c.name.startswith("drawing") and c.name.endswith(".svg")
        ][0]
        bpy.ops.object.select_all(action="DESELECT")
        if not drawing_col.objects:
            return False

        for obj in drawing_col.objects:
            obj.select_set(True)

        context.view_layer.objects.active = drawing_col.objects[0]
        bpy.ops.object.join()
        mesh = bpy.data.meshes.new_from_object(
            drawing_col.objects[0],
            preserve_all_data_layers=False,
            depsgraph=context.evaluated_depsgraph_get(),
        )

        bpy.data.curves.remove(drawing_col.objects[0].data)
        bpy.data.collections.remove(drawing_col)

        drawing_obj = bpy.data.objects.new("drawing", mesh)
        self.drawings_collection.objects.link(drawing_obj)

        # apply scale
        prev_scale = self.object.scale.copy()
        self.apply_scale(context)

        drawing_obj.parent = plane.object
        offset = plane.camera.data.ortho_scale / 2
        drawing_obj.location = Vector((-offset, offset, 0))
        scale = offset * 7.088
        drawing_obj.scale = Vector((scale, scale, 0))

        context.view_layer.update()
        for v in drawing_obj.data.vertices:
            v.co = drawing_obj.matrix_world @ v.co

        drawing_obj.parent = None
        drawing_obj.location = Vector((0, 0, 0))
        drawing_obj.rotation_euler = Vector((0, 0, 0))
        drawing_obj.scale = Vector((1, 1, 1))

        # gpencil object
        gpencil_mesh_obj = drawing_obj.copy()
        gpencil_mesh_obj.data = drawing_obj.data.copy()
        self.drawings_collection.objects.link(gpencil_mesh_obj)

        context.view_layer.objects.active = gpencil_mesh_obj
        gpencil_mesh_obj.select_set(True)
        bpy.ops.object.convert(
            target="GPENCIL", keep_original=False, thickness=3, offset=0
        )

        drawing_obj.hide_render = True
        drawing_obj.hide_viewport = True
        drawing_obj.hide_select = True

        # transform drawing_obj into xy plane for export
        for vert in drawing_obj.data.vertices:
            vert.co = plane.object.matrix_world.inverted() @ vert.co

        drawing_obj.location = Vector((0, 0, 0))

        # recover scale
        self.recover_scale(context, prev_scale)

        gpencil_obj = context.view_layer.objects.active
        prev_col = [
            c
            for c in list((context.scene.collection,))
            + list(context.scene.collection.children)
            if gpencil_obj in c.objects.values()
        ]
        if prev_col:
            try:
                prev_col[0].objects.unlink(gpencil_obj)
            except Exception as e:
                print(f"Could not unlink GPencil object: {e}")

        self.drawings_collection.objects.link(gpencil_obj)
        gpencil_obj.hide_select = True

        return drawing_obj, gpencil_obj

    def apply(self, context, copy):
        [o.select_set(False) for o in context.view_layer.objects.selected]
        if copy:
            for obj in self.iter_affected_objects():
                new_obj = obj.copy()
                new_obj.data = obj.data.copy()
                new_obj.name = f"{self.name}.{obj.name}"
                context.scene.collection.objects.link(new_obj)
                context.view_layer.objects.active = new_obj
                new_obj.select_set(True)
                mod = new_obj.modifiers.get(f"{self.name}.boolean")
                if mod:
                    bpy.ops.object.modifier_apply("INVOKE_DEFAULT", modifier=mod.name)
        else:
            for obj in self.iter_affected_objects():
                context.view_layer.objects.active = obj
                obj.select_set(True)
                mod = obj.modifiers.get(f"{self.name}.boolean")
                if mod:
                    bpy.ops.object.modifier_apply("INVOKE_DEFAULT", modifier=mod.name)

            self.delete(context)
