import math
from mathutils import Vector
from functools import cache

import bpy
from bpy.types import PropertyGroup


from ..utils.items import material_items
from ..utils.decorators import update_decorator


@update_decorator
def update_invert(self, context):
    operation = "DIFFERENCE" if self.invert else "INTERSECT"
    for obj in self.iter_affected_objects():
        mod = obj.modifiers.get(f"{self.name}.boolean")
        if mod and mod.operation != operation:
            mod.operation = operation

    for mat in self.iter_affected_materials():
        n = mat.node_tree.nodes.get(self.name)
        if n:
            n.inputs[2].default_value = 1 if self.invert else 0

    # update cameras
    for plane in self.planes:
        plane.update_invert(self.invert)


@update_decorator
def update_enable(self, context):
    update_appearance(self, context)


@update_decorator
def update_appearance(self, context):
    for obj in self.iter_affected_objects():
        mod = obj.modifiers.get(f"{self.object.name}.boolean")
        if mod:
            show = self.method == "BOOLEAN"
            show_viewport = show and self.enable
            if mod.show_viewport != show_viewport:
                mod.show_viewport = show_viewport
            show_render = show and self.enable and not self.hide_render
            if mod.show_render != show_render:
                mod.show_render = show_render

    for mat in self.iter_affected_materials():
        n = mat.node_tree.nodes.get(self.object.name)
        if n:
            show = self.method == "SHADER" and self.enable
            n.mute = not show


@update_decorator
def update_hide(self, context):
    self.object.hide_set(self.hide)


@update_decorator
def update_hide_render(self, context):
    update_appearance(self, context)


@update_decorator
def update_material(self, context):
    if not self.cutter:
        return

    self.cutter.data.materials.clear()
    if self.material != "From Mesh":
        material = bpy.data.materials.get(self.material)
        if material:
            self.cutter.data.materials.append(material)


@update_decorator
def update_color(self, context):
    for mat in self.iter_affected_materials():
        n = mat.node_tree.nodes.get(self.name)
        if n:
            n.inputs[1].default_value = Vector(self.color).to_4d()


@update_decorator
def update_method(self, context):
    update_appearance(self, context)


@update_decorator
def update_boolean_settings(self, context):
    sb = self.get_sb(context)
    if not sb:
        return

    for obj in sb.iter_affected_objects():
        mod = obj.modifiers.get(f"{self.name}.boolean")
        if mod:
            if mod.use_self != self.boolean_use_self:
                mod.use_self = self.boolean_use_self
            if mod.use_hole_tolerant != self.boolean_use_hole_tolerant:
                mod.use_hole_tolerant = self.boolean_use_hole_tolerant


# drawing update methods
def update_cross_section(self, context, prop, planes=None):
    if not planes:
        sb = getattr(self, ("get_sb"), lambda c: None)(context)
        if not sb:
            return

        planes = sb.planes

    value = getattr(self, f"cross_section_{prop}")
    if getattr(self, f"elevation_{prop}_link"):
        update_elevation(self, context, prop, planes)

    for plane in planes:
        obj = getattr(plane, "cross_section_gpencil", None)
        if not obj:
            continue
        globals()[f"set_{prop}"](self, context, plane, obj, value)


def update_elevation(self, context, prop, planes=None):
    if not planes:
        sb = getattr(self, ("get_sb"), lambda c: None)(context)
        if not sb:
            return

        planes = sb.planes

    if getattr(self, f"elevation_{prop}_link"):
        value = getattr(self, f"cross_section_{prop}")
    else:
        value = getattr(self, f"elevation_{prop}")

    for plane in planes:
        obj = getattr(plane, f"elevation_gpencil", None)
        if not obj:
            continue
        globals()[f"set_{prop}"](self, context, plane, obj, value)


def set_dst(self, context, plane, obj, dst):
    if self.invert:
        dst = -dst

    face = self.object.data.polygons[plane.id]
    mw = self.object.matrix_world
    normal = mw @ face.normal - mw @ Vector((0, 0, 0))
    obj.location = dst * normal


def set_thickness(self, context, plane, obj, thickness):
    obj.data.pixel_factor = thickness


def set_color(self, context, plane, obj, color):
    tint = Vector(color[:3])
    obj.data.layers.active.tint_color = tint
    obj.data.layers.active.opacity = color[3]


class SECTIONBOX_PG_SectionBoxSettings(PropertyGroup):
    update_: bpy.props.BoolProperty()

    invert: bpy.props.BoolProperty(
        name="Invert",
        update=update_invert,
    )
    enable: bpy.props.BoolProperty(
        name="Enable",
        default=True,
        update=update_enable,
    )
    hide: bpy.props.BoolProperty(
        name="Hide",
        update=update_hide,
    )
    hide_render: bpy.props.BoolProperty(
        name="Hide Render",
        update=update_hide_render,
    )
    material: bpy.props.EnumProperty(
        name="Material",
        items=material_items,
        update=update_material,
    )
    color: bpy.props.FloatVectorProperty(
        name="Color",
        size=3,
        default=(0.0, 0.0, 0.0),
        subtype="COLOR",
        min=0.0,
        max=1.0,
        update=update_color,
    )
    method: bpy.props.EnumProperty(
        name="Method",
        items=[
            ("SHADER", "Shader", "", "SHADING_TEXTURE", 0),
            ("BOOLEAN", "Boolean", "", "MOD_BOOLEAN", 1),
        ],
        update=update_method,
    )
    boolean_use_self: bpy.props.BoolProperty(
        name="Self Intersection",
        description="Makes the section working with self-intersecting meshes",
        default=False,
        update=update_boolean_settings,
    )
    boolean_use_hole_tolerant: bpy.props.BoolProperty(
        name="Hole Tolerant",
        description="Makes the section working with meshes with holes",
        default=True,
        update=update_boolean_settings,
    )

    # cross section settings
    auto_cross_section: bpy.props.BoolProperty(
        name="Auto Cross Section",
        description="Automatically create cross sections for this plane",
        default=False,
    )
    cross_section_dst: bpy.props.FloatProperty(
        name="Cross Section Distance",
        description="Distance from the section plane to the cross section",
        default=0.5,
        soft_min=0,
        update=lambda s, c: update_cross_section(s, c, "dst"),
    )
    cross_section_entire: bpy.props.BoolProperty(
        name="Cross Section Entire",
        description="Create cross section from the entire mesh instead of the visible part\n(Update the cross section to apply this setting)",
        default=False,
    )

    cross_section_color: bpy.props.FloatVectorProperty(
        name="Cross Section Color",
        description="Color of the cross section",
        subtype="COLOR",
        size=4,
        default=(0.0, 0.0, 0.0, 1.0),
        min=0,
        max=1,
        update=lambda s, c: update_cross_section(s, c, "color"),
    )

    cross_section_thickness: bpy.props.FloatProperty(
        name="Cross Section Thickness",
        description="Thickness of the cross section lines",
        default=1,
        soft_min=0.0,
        update=lambda s, c: update_cross_section(s, c, "thickness"),
    )

    # elevation settings
    elevation_create_link: bpy.props.BoolProperty(
        name="Elevation Link Creation",
        description="Link elevation creation to cross section creation",
        default=False,
    )
    elevation_crease_angle: bpy.props.FloatProperty(
        name="Elevation Crease Angle",
        description="Edge angle threshold for generating the elevation\n(Update the elevation to apply this setting)",
        default=(math.pi / 180) * 100,
        subtype="ANGLE",
        min=0.0,
        max=math.pi,
    )
    elevation_dst: bpy.props.FloatProperty(
        name="Elevation Distance",
        description="Distance from the section plane to the elevation",
        default=0.5,
        soft_min=0.0,
        update=lambda s, c: update_elevation(s, c, "dst"),
    )
    elevation_dst_link: bpy.props.BoolProperty(
        name="Elevation Link Distance",
        description="Link elevation distance to cross section distance",
        default=True,
        update=lambda s, c: update_elevation(s, c, "dst"),
    )
    elevation_thickness: bpy.props.FloatProperty(
        name="Elevation Thickness",
        description="Thickness of the elevation lines",
        default=0.5,
        soft_min=0.0,
        update=lambda s, c: update_elevation(s, c, "thickness"),
    )
    elevation_thickness_link: bpy.props.BoolProperty(
        name="Elevation Link thickness",
        description="Link elevation thickness to cross section thickness",
        default=True,
        update=lambda s, c: update_elevation(s, c, "thickness"),
    )
    elevation_color: bpy.props.FloatVectorProperty(
        name="Elevation Color",
        description="Color of the elevation",
        subtype="COLOR",
        size=4,
        default=(0.0, 0.0, 0.0, 1.0),
        min=0.0,
        max=1.0,
        update=lambda s, c: update_elevation(s, c, "color"),
    )
    elevation_color_link: bpy.props.BoolProperty(
        name="Elevation Link color",
        description="Link elevation color to cross section color",
        default=True,
        update=lambda s, c: update_elevation(s, c, "color"),
    )

    @cache
    def get_sb(self, context):
        if not self.name:
            return

        sb = context.scene.sectionbox.list.get(self.name)
        return sb.object.sectionbox if sb else None
