import requests

import bpy
from bpy.types import Operator

from .utils.preferences import get_prefs
from . import addon_id

bl_info = {
    "name": "Section Box",
    "description": "Visualization, Sections, Elevations, and more vfxMed",
    "blender": (3, 4, 0),
    "version": (2, 1, 2),
    "category": "3D View",
    "author": "Ruben Messerschmidt",
    "location": "View3d > Sidebar > Section Box",
    "doc_url": "https://section-box.readthedocs.io/en/latest/",
    "version_url": "https://raw.githubusercontent.com/rubenmesserschmidt/Section-Box_docs/main/version",
}


show_popup = True


def invoke_popup():
    global show_popup
    if not show_popup:
        return

    show_popup = False
    getattr(getattr(bpy.ops, "sectionbox"), "update_check")("INVOKE_DEFAULT")


def check_for_updates():
    try:
        data = requests.get(
            bl_info["version_url"],
        )

    except:
        return -1

    if data.status_code == 200:
        latest_version = tuple(int(n) for n in (data.text.split(".")))
        if latest_version > bl_info["version"]:
            return latest_version

    return False


class UpdateCheck(Operator):
    bl_idname = "sectionbox.update_check"
    bl_label = ""
    bl_description = "Check for updates (requires internet connection)"
    bl_options = {"INTERNAL"}

    manual: bpy.props.BoolProperty(
        options={"HIDDEN"}
    )  # called manually from addon preferences
    ignore_update: bpy.props.BoolProperty(options={"HIDDEN"})

    def invoke(self, context, event):
        self.prefs = get_prefs()

        if not self.prefs.update_check and not self.manual:
            return {"CANCELLED"}

        self.mouse_co = (event.mouse_x, event.mouse_y)

        self.update = check_for_updates()
        if self.update == -1:
            self.report(
                {"ERROR"},
                "Could not connect to server, please check your internet connection!",
            )
            return {"FINISHED"}
        elif self.update == tuple(self.prefs.ignore_update):
            if self.manual:
                self.ignore_update = True
            else:
                return {"CANCELLED"}

        if self.update:
            context.window.cursor_warp(
                int(context.window.width / 2), int(context.window.height / 2)
            )
            self.mouse_reset = True
            return context.window_manager.invoke_props_dialog(self, width=300)

        else:
            self.report({"INFO"}, f'{bl_info["name"]} is up to date!')
            return {"FINISHED"}

    def execute(self, context):
        if self.ignore_update:
            self.prefs.ignore_update = self.update
            self.report({"INFO"}, f"Ignoring update")
        else:
            self.prefs.ignore_update = (0, 0, 0)

        bpy.ops.wm.save_userpref()
        return {"FINISHED"}

    def draw(self, context):
        if self.mouse_reset:
            context.window.cursor_warp(int(self.mouse_co[0]), int(self.mouse_co[1]))
            self.mouse_reset = False

        layout = self.layout
        row = layout.row()
        row.alignment = "CENTER"
        row.scale_y = 0.3
        row.label(text=f"A new version of {bl_info['name']} is available!")

        row = layout.row()
        row.alignment = "CENTER"
        row.label(
            text=f"Explore whats new in {self.update[0]}.{self.update[1]}.{self.update[2]}"
        )

        # update resources
        row = layout.row(align=True)
        row.alignment = "CENTER"
        row.operator(
            "wm.url_open", text="Update Guide", icon="HELP"
        ).url = f'{bl_info["doc_url"]}update.html'
        row.operator(
            "wm.url_open", text="See Changelog", icon="TEXT"
        ).url = f'{bl_info["doc_url"]}changelog.html'
        layout.separator()

        # update check
        if not self.manual:
            row = layout.row()
            row.alignment = "CENTER"
            row.prop(self.prefs, "update_check", text="Check for updates on startup")
        layout.separator()

        # ignore update
        row = layout.row()
        row.scale_x = 3
        row.alignment = "CENTER"
        row.prop(self, "ignore_update", text="Ignore this update", toggle=True)
